#!/usr/bin/python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

def new_fun(arg1 = 'Default', arg2 = 'Default'):
    """Short explanation.

    This is the very long explanation

    Parameters
    -----------
    arg1 : type, default : 'Default'
        What is arg1.

    arg2 : {'Default', 'Different', 'Another'}
        What is arg2.

    Returns
    -----------
    out : type
        What is out.
    """
